/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: mp3_file.cc 729 2007-05-20 14:55:00Z  $

#include "mp3_file.h"

#include <iostream>
#include <cassert>
#include "str_utils.h"
#include "spinner.h"
#include "config.h"

// File structure based on <http://www.multiweb.cz/twoinches/MP3inside.htm>

namespace net_outlyer {

namespace mp3 {

namespace str=net_outlyer::str_utils;

using std::ifstream;
using std::cerr;
using std::cout;
using std::string;
using std::ios;

// -- MP3 File --

void mp3file::check_file_type(ifstream & in) throw (e_error) {
    in.seekg(0, std::ios::beg);

    uint8_t filestart[3];
    in.read(CHARR(filestart), 3);

    if (0 == memcmp(filestart, "ID3", 3)) {
        file_type = FT_ID3V2;
    }
    else if (filestart[0] == 0xff && 0xf0 == (filestart[1] & 0xf0)) {
        file_type = FT_MP3;
    }
    else {
        throw e_bad_filetype();
    }
}

void mp3file::read_info(ifstream & in) throw (e_error) {
    check_file_type(in);

    in.seekg(0, std::ios::beg);
    if (FT_ID3V2 == file_type) {
        id3v2.read(in);

        // Jump to tag end:
        in.seekg(id3v2.tag_size, std::ios::cur); // ID3v2 total size is header(=10) + tag_size
    }

    // Find first frame
    seek_to_next_frame(in); // throws e_read_failure if none found
    // POST ]-> at first frame

    std::auto_ptr<mp3frame_first> fframe (new mp3frame_first(meta_info));
    fframe->read(in);
    fframe->frame_idx = 1;
    frames.push_back( *fframe );

    net_outlyer::spinner spinner;
    try {
        uint32_t frame_idx = 2;

        spinner.start();

        while (true) {
            seek_to_next_frame(in);
            mp3frame f;
            f.frame_idx = frame_idx++;
            f.read(in);
            f.skip_data(in);
            frames.push_back(f);

            spinner.step();
        }
    }
    catch (e_read_failure & e) {
        spinner.stop();
        cerr << "No more frames found after byte #" << frames.back().start << std::endl;
    }
    spinner.stop();
    cerr << frames.back().frame_idx << " frames read" << std::endl;
}

void mp3file::dump_meta_info() const {
    if (FT_ID3V2 == file_type) {
        id3v2.dump();
    }
    else {
        cout << "This file doesn't contain an ID3v2 header" << std::endl;
    }
    frames[0].dump();

    cout << "VBR: " << (meta_info.is_vbr ? "Yes\n" : "No\n");
    if (meta_info.is_vbr) {
        if (meta_info.has_numframes) {
            cout << "\tNumber of frames included (" << meta_info.num_frames << ")\n";
        }
        if (meta_info.has_filesize)
            cout << "\tFile size included (" << meta_info.file_size << ")\n";
        if (meta_info.has_toc)
            cout << "\tTOC included\n";
        if (meta_info.has_vbr_scale)
            cout << "\tVBR scale included\n";
    }
}

void mp3file::plot() const {
    uint32_t dist[0x10];
    memset(dist, 0x00, sizeof(dist));

    uint32_t max = 0;

    {
        // The first frame *on VBR files* doesn't contain audio. It just contains meta info.
        std::vector<mp3frame>::const_iterator it=frames.begin();
        if (meta_info.is_vbr) {
            ++it;
        }
        for ( ; it != frames.end(); ++it)
        {
            const mp3frame & f = *it;
            // index == 0x10 means invalid bitrate
            const uint16_t idx = mp3element::bitrate_index(f.bitrate);
            assert( idx < 0x10 );
            dist[idx]++;
            if (max < dist[idx]) max=dist[idx];
        }
    }

    uint32_t sum=0, fs=0;

    // Max length of bitrate tags
    static const size_t LBR = string("invalid").length();
    // Max length of number of frames, + padding
    const size_t LOC = str::to_string(max).length() + 1;

    // Console width
    // FIXME: detect
    const uint16_t CONWIDTH = 80;
    // 4 = "[]: " + 1 => space between number of frames and dots
    const uint16_t AVAIL = CONWIDTH - (LBR + LOC + 5);

    float sumbrs = 0;

    for (int i=0; i<0x0f; ++i) {
        string bitrate = str::to_string(BITRATE_TABLE[i]) + "kbps";
        string occ = str::to_string(dist[i]);

        cout << "[" << str::pad(bitrate, LBR, " ", str::PAD_LEFT) << "]: "
             << str::pad(occ, LOC);

        // the scale is relative to the max
        const int SCALED = int(dist[i] * static_cast<float>(AVAIL) / max);

        cout << str::pad("", SCALED, PLOT_SYMBOL);

        cout << "\n";
        sum += dist[i]; // dist[i] = number of frames of this bitrate
        sumbrs += (dist[i] * BITRATE_TABLE[i]);
        // That's the number of bits per seconds at this bitrate
        const uint32_t secbits = dist[i] * BITRATE_TABLE[i] * 1000;
        // Each frame is 0.026 seconds long
        fs+=int( secbits * 0.026);
    }
    cout << "[invalid]: " << dist[0x0f] << "\n";

    string abr = str::to_string(sumbrs/sum);
    const string::size_type dotidx = abr.find_first_of('.');
    if (string::npos != dotidx) {
        if ((dotidx + 3) < abr.length()) {
            abr.erase(dotidx+3);
        }
    }

    cout << "Average bitrate is " << abr << " kbps\n";
    // FIXME: There seems to be a 1 frame difference
    cout << "\twith a total of " << sum << " frames" << std::endl;
    cout << "\twhich yields a theoretical (audio data) size of "
         << (fs / 8) << " bytes" << std::endl;
}

void mp3file::seek_to_next_frame(ifstream & in) throw (e_read_failure) {
    uint8_t last=0, byte;

    // Frames start with FF Fx

    while (in.good()) {
        in.read(CHARR(&byte), 1);

        if ((last==0xff) && 0xf0 == (byte&0xf0)) {
            in.seekg( -2 , std::ios::cur);
            return;
        }

        last = byte;
    }

    throw e_read_failure("Couldn't find next frame.", EX_DATAERR);
}

} // namespace mp3

} // namespace net_outlyer

/* vim:set ts=4 et ai: */
